/***************************************************************************
 *                                                                         *
 *                  (begin: Feb 20 2003)                                   *
 *                                                                         *
 *   Parallel IQPNNI - Important Quartet Puzzle with NNI                   *
 *                                                                         *
 *   Copyright (C) 2005 by Le Sy Vinh, Bui Quang Minh, Arndt von Haeseler  *
 *   Copyright (C) 2003-2004 by Le Sy Vinh, Arndt von Haeseler             *
 *   {vinh,minh}@cs.uni-duesseldorf.de                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef LIBR_H
#define LIBR_H

#include "br.h"
#include "lind.h"

class LiBr : public Br<double> {
public:
	//the constructor
	LiBr ();

	//smaLiNd, smaNdNo_ ===================greNdNo, greLiNd
	LiNd *smaLiNd_;
	LiNd *greLiNd_;

	double score_;
	int isSwapLChi_;
	int isCan_, isSwap_;
	void writeInf ();

	//copy the base br into this likelihood branch
	void copyBase (Br<double> &br);

	void setTailLiNd (LiNd *newTailLiNd);

	//get the remaining liNd of this branch
	LiNd &getRemLiNd (int existNdNo);

	//get the sma liNd of this branch
	LiNd &getSmaLiNd ();

	//get the gre liNd of this branch
	LiNd &getGreLiNd ();

	//get the tail liNd of this branch
	LiNd &getTailLiNd ();

	//get the head liNd of this branch
	LiNd &getHeadLiNd ();

	LiNd &getLiNd (int ndNo);

	int isOutGrp ();

	//create the pattern for all liNds of this branch
	void createExDesLiNd ();

	//set the two liNds to be NOT cmped
	void openLiNd ();

	//set the tail liNd to be Not cmped
	void openTailLiNd ();

	//set the tail liNd to be Not cmped
	void openHeadLiNd ();

	//compute the likelihood for liNd of this branch
	void cmpLiNd (int ndNo);

	//compute the likelihood for tail liNd of this branch
	void cmpTailLiNd ();

	//compute the likelihood for head liNd of this branch
	void cmpHeadLiNd ();

	//reCompute the likelihood for tail liNd of this branch
	void reCmpTailLiNd ();

	//reCompute the likelihood for head liNd of this branch
	void reCmpHeadLiNd ();

	//compute the likelihood along this branch
	double cmpLi (int ptnNo, int rateNo, int parBaseNo, int parNdNo);

	/***********************************************************************
	* compute the probalibity change matrix times base frequency
	***********************************************************************/	
	
	//cmp the vector probability of change from baseNo
	//void cmpProbChangeFrq (const int baseNo, const double brLen, DVec20 &probVec);

	/**
	cmp the array of probability of changing from
	one nucleotide into one nucleotide after a period of brLen / subRate_
	*/
	void cmpProbChangeFrq (const double brLen, DMat20 &probMat);
	
	/**
	cmp the array of probability of changing from
	one nucleotide into one nucleotide after a period of brLen / subRate_
	*/
	void cmpProbChangeFrq (const double brLen, const int classNo, DMat20 &probMat);

	//virtual void cmpProbChangeDerivatives (const int stateNo1, const int stateNo2, const double brLen, double &derv1, double &derv2);

	//void cmpProbChangeDerivativesFrq (const int stateNo1, const double brLen, DVec20 &probVec, DVec20 &derv1, DVec20 &derv2);
	
	void cmpProbChangeDerivativesFrq (const double brLen, DMat20 &probMat, DMat20 &derv1, DMat20 &derv2);
	
	void cmpProbChangeDerivativesFrq (const double brLen, int classNo, DMat20 &probMat,
        DMat20 &derv1, DMat20 &derv2);
	/***********************************************************************
	* compute likelihood from INTERNAL branch
	***********************************************************************/
	
	//-----------------------------------------------
	//-  UNIFORM rate	
	//-----------------------------------------------
	/**
		compute the log likelihood of this tree via this an internal branch
	*/
	double cmpInLogLiUniformRate ();

	/**
		compute the log likelihood derivative via internal branch
	*/
	double cmpInLogLiDerivativeUniformRate (double &logli_derv1, double &logli_derv2, bool calc_logli);
	
	//-----------------------------------------------
	//-  SITE-SPECIFIC rate	
	//-----------------------------------------------
	
	/**
		compute the log likelihood of this tree via this an external branch
	*/
	double cmpInLogLiSpecificRate ();

	/**
		compute the log likelihood of this tree via this an external branch
	*/
	double cmpInLogLiDerivativeSpecificRate (double &logli_derv1, double &logli_derv2, bool calc_logli);
	
	
	//-----------------------------------------------
	//-  GAMMA rate	
	//-----------------------------------------------
	
	/**
		compute the log likelihood with gamma rate of this tree via this an external branch
	*/
	double cmpInLogLiGammaRate ();

	/**
		compute the log likelihood with gamma rate of this tree via this an external branch
	*/
	double cmpInLogLiDerivativeGammaRate (double &logli_derv1, double &logli_derv2, bool calc_logli);
	
	
	
	//-----------------------------------------------
	//-  codon-based Nielsen Yang Model
	//-----------------------------------------------
	/**
		compute the log likelihood of this tree via this an internal branch
	*/
	double cmpInLogLiSelection ();

	/**
		compute the log likelihood derivative via internal branch
	*/
	double cmpInLogLiDerivativeSelection (double &logli_derv1, double &logli_derv2, bool calc_logli);
	
	
	/***********************************************************************
	* compute likelihood from EXTERNAL branch
	***********************************************************************/
	
	//-----------------------------------------------
	//-  UNIFORM rate	
	//-----------------------------------------------
	
	/**
		compute the log likelihood of this tree via this an external branch in case of one rate
	*/
	double cmpExLogLiUniformRate ();
	
	/**
		compute the log likelihood of this tree via this an external branch in case of one rate
	*/
	double cmpExLogLiDerivativeUniformRate (double &logli_derv1, double &logli_derv2, bool calc_logli);


	//-----------------------------------------------
	//-  SITE-SPECIFIC rate	
	//-----------------------------------------------
	
	/**
		compute the log likelihood of this tree via this an external branch in case of one rate
	*/
	double cmpExLogLiSpecificRate ();

	/**
		compute the log likelihood of this tree via this an external branch in case of one rate
	*/
	double cmpExLogLiDerivativeSpecificRate (double &logli_derv1, double &logli_derv2, bool calc_logli);
	
	
	//-----------------------------------------------
	//-  GAMMA rate	
	//-----------------------------------------------
	
	/**
		compute the log likelihood of this tree via this an external branch in case of one rate
	*/
	double cmpExLogLiGammaRate ();

	/**
		compute the log likelihood of this tree via this an external branch in case of one rate
	*/
	double cmpExLogLiDerivativeGammaRate (double &logli_derv1, double &logli_derv2, bool calc_logli);
	
	//-----------------------------------------------
	//-  codon-based Nielsen Yang Model
	//-----------------------------------------------
	
	/**
		compute the log likelihood of this tree via this an external branch in case of one rate
	*/
	double cmpExLogLiSelection ();
	
	/**
		compute the log likelihood of this tree via this an external branch in case of one rate
	*/
	double cmpExLogLiDerivativeSelection (double &logli_derv1, double &logli_derv2, bool calc_logli);

	
	
	/***********************************************************************
	* general computing likelihood 
	***********************************************************************/
	
	// compute the log-likelihood from external branch
	double cmpExLogLi ();

	// compute the log-likelihood from internal branch
	double cmpInLogLi ();

	/**
		 compute the log-likelihood from internal branch
	*/
	double cmpInLogLiDerivatives (double &logli_derv1, double &logli_derv2, bool calc_logli);
	
	
	/** 
		compute the log-likelihood derivatives from external branch
	*/
	double cmpExLogLiDerivatives (double &logli_derv1, double &logli_derv2, bool calc_logli);
	
	
#ifdef PARALLEL

	// parallel compute the log likelihood of this tree via this an external branch
	double par_cmpInLogLiUniformRate ();

	// parallel compute the log likelihood of this tree via this an external branch
	double par_cmpInLogLiSpecificRate ();

	// parallel compute the log likelihood of this tree via this an external branch
	double par_cmpInLogLiGammaRate ();

	// parallel compute the log likelihood of this tree via this an external branch in case of one rate
	double par_cmpExLogLiUniformRate ();

	// parallel compute the log likelihood of this tree via this an external branch in case of one rate
	double par_cmpExLogLiSpecificRate ();

	//parallel compute the log likelihood of this tree via this an external branch in case of one rate
	double par_cmpExLogLiGammaRate ();

#endif

	void empiricalBayesAnalysis();
	//set memory for this liBr
	void setLimit ();

	//release all memory of this class

	void release ();

	//the destructor
	virtual ~LiBr ();
};
#endif
