/***************************************************************************
 *                                                                         *
 *                  (begin: Feb 20 2003)                                   *
 *                                                                         *
 *   Parallel IQPNNI - Important Quartet Puzzle with NNI                   *
 *                                                                         *
 *   Copyright (C) 2005 by Le Sy Vinh, Bui Quang Minh, Arndt von Haeseler  *
 *   Copyright (C) 2003-2004 by Le Sy Vinh, Arndt von Haeseler             *
 *   {vinh,minh}@cs.uni-duesseldorf.de                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef OPTURTREE_H
#define OPTURTREE_H

#include "exndarr.h"
#include "inndarr.h"
#include "brarr.h"
#include "libr.h"
#include "lind.h"
#include "librarr.h"
#include "vec.h"
#include "dmat20.h"
#include "brent.h"

const bool PARALLEL_LOGLI = true;
const int DO_PARALLEL_LOGLI_FUNC = 1;
const int DO_PARALLEL_OTHER_FUNC = 2;

/**
this is one of the most important class, it describles all things involving
optimizing an unrooted tree */



class OptUrTree : public Brent {
public:
	friend class LiNd;
	friend class LiBr;

	//the constructor function
	OptUrTree();

	//use this function for constructing this class
	void doConstructor ();

	void changeNd (int parNdno, int sNdno, int chiNdNo, int dNdNo);

	double optLocalBr (int brNo);

	//int optTopology (int parNdNo, int sNdNo, int dNdNo);

	void sortScoreBr ();
	/**
		sort the branch scores (for NNI) by Quicksort
	*/
	void sortScoreBr (int left, int right);

	int findNPosSwapBr (int maxNSwapBr);
	void trySwapBr (double lamda);
	int swapBr ();
	void swapBrBack ();
	double cmpSwapLogLi ();
	double optSwapLogLi ();
	void createScoreBr ();
#ifdef PARALLEL
	void par_createScoreBr ();
	void buildDataType();
#endif // PARALLEL
	/**
		optimize the tree topology with NNI
	*/
	double optTopology (bool &topo_changed);
	void draw ();

	double oriLenArr_[MAX_NUM_SEQ * 2];
	double canLenArr_[MAX_NUM_SEQ * 2];
	double optedLenArr_[MAX_NUM_SEQ * 2];

	//set out group node no for this tree
	void setOutGrpNd (int outGrpNdNo);

	//set the exNdArr for this class, construct the structure of this unrooted tree
	void setExNd (ExNdArr &exNdArr);

	//set the inNdArr for this class, construct the structure of this unrooted tree
	void setInNd (InNdArr &inNdArr);

	//set the brArr for this class, construct the structure of this unrooted tree
	void setBr (BrArr<double> &brArr);

	//get the brArr of this tree after optimizing
	void getBr (BrArr<double> &brArr);

	//get the exNdArr of this tree after optimizing
	void getExNd (ExNdArr &ExNdArr);

	//get the InNdArr of this tree after optimizing
	void getInNd (InNdArr &InNdArr);

	double getBrLen (int brNo);

	// optimize all branch length of the tree
	double optBranches(int nIt = MAX_IT_UR_BR);


	//opt all childrent bres of this node
	void optBranches(int ndNo, int &nChangedBr);


	//clean all content of this class
	void clean();

	//everything is inited here
	void init ();

	//return the log likelihood of this tree.
	double getLogLi ();

	//write all information of this optimal unrooted tree into the screen
	void write ();

	void updateBestLogLi (double logLi);

	//release all memmory for this class

	void release ();

	//the destructor of this class
	virtual ~OptUrTree();

	/***********************************************************************************
	***********************************************************************************
	***********************************************************************************
	 
	***********************************************************************************
	***********************************************************************************/

	//private :

	Vec<int> inBrNoLs_;

	double bestLogLi_;

	//the current number of inNds in this tree
	int curNInNd_;

	//the current number of ExNds in this tree
	int curNExNd_;

	//the current number of bres in this tree
	int curNBr_;

	//the maximum number of inNds in this tree
	int maxNInNd_;

	//the maximum number of ExNds in this tree
	int maxNExNd_;

	//the maximum number of bres in this tree
	int maxNBr_;

	//the outGrpNdNo if this unrooted tree
	int outGrpNdNo_;


	//the outGrpBrNo if this unrooted tree
	int outGrpBrNo_;


	//the artificial internal root node if this unrooted tree
	int artInRootNo_;

	//the array of all internal nds of this UnrootTree
	InNdArr inNdArr_;

	//the array of all ex nds of this UnrootTree
	ExNdArr exNdArr_;

	//the array of all bres of this UnrootTree
	LiBrArr liBrArr_;

	//isLiCmped = 0, it means that the likelihood of this tree has NO been computed
	int isLiCmped_;


	//isBrLenEsted = 1, it means that we have estimated the len of all branches
	int isBrLenEsted_;

	//contains the branch no which is changed
	int actBrNo_;

	//it contains the genetic distance between any two sequences
	Mat<double> *genDisMat_;

	//the number of calling function cmpNegLogLiBr
	int nCmpNegLogLiBr_;

	//the number of compute likehood of an internal node
	int nCmpLi_;

	int nCmpExLogLi_;
	int nCmpInLogLi_;

	//the number of calling cmpLogLi
	int nCmpLogLi_;

	int isInNd (int ndNo);

	//check whether or not it is a external node
	int isExNd (int ndNo);

	//create exDesLiInNd for all internal nodes of this tree
	void createExDesLiNd ();

	//create pattern for all internal node of this tree
	void createPtnLiNd ();

	/*create the list of all exnternal nodes which belong to the same one root */
	void createGrpExNd (const int parGrpNo, const int grpNo, Vec<int> &grpExNdNoLs);


	/*compute the distance between two groups, group 1 and group 2.
	group1 is a subtree whose the root is grpNo1
	group2 is a subtree whose the root is grpNo2
	parentNo is the parent of grpNo1, and grpNo2
	*/
	double cmpGrpDis (const int parGrpNo, const int grpNo1, const int grpNo2);
	double cmpGrpDis (Vec<int> &grpExNdNoLs1, Vec<int> &grpExNdNoLs2);
	double cmpGrpDis (int exNdNo, Vec<int> &grpExNdNoLs);


	/*using Rzhetsky and Nei method to estimate the branch of this tree as
	a starting point for maximum likelihood*/
	double estExBrLen (Br<double> &exBr);

	/*using Rzhetsky and Nei method to estimate the branch of this tree as
	a starting point for maximum likelihood*/
	double estInBrLen (Br<double> &inBr);

	/*using Rzhetsky and Nei method to estimate the branch of this tree as
	a starting point for maximum likelihood*/
	void estBrLen ();


	//assign values for all LiNd_ of all LiNd
	void cmpLiNd ();

	//assign values for all LiNd_ of all LiNd
	void cmpHeadLiNd ();

	//compute the likelihood at a given pattern site

	LDOUBLE cmpPtnLi (int ptnNo_);


	//cmp the log likelihood of this tree
	double cmpLogLi ();


	//serve for cmpLiBr and cmpLogLiBr
	void changeBr (double len);

	//cmp the negative log likelihood of this tree when only this br changes
	//using for optimizing by brent method
	virtual double cmpNegLogLi (double len);

	double cmpLogLiDerivatives (double brLen, double &logli_derv1, double &logli_derv2, bool calc_logli);



	//optimize the br whose name is brNo, return the length of br
	double optBr (int brNo);


	/**
		Newton-Raphson method
	*/
	double doNewtonRaphson(double guess, double &logli);

#ifdef PARALLEL
	void turnOnParallelPhase(bool par_logli);
	void turnOffParallelPhase();
	bool isParallelPhase(bool par_logli);
#endif // PARALLEL

private:
	/**
		sum of score in topology optimization
	*/
	double score_sum, original_logli, upper_bound_logli;


	int do_parallel_phase;

}
; //end of class optimizing unrooted tree

extern OptUrTree opt_urtree;

#endif // OPTIMALUNROOTTREE


